#ifndef N_VAR_H
#define N_VAR_H
//------------------------------------------------------------------------------
/**
    @class nVar

    @contains a piece of data
        either an int, float, vector3, quaternion, string, ref<nThing> or a void*

    (C)	2004	Chris Miles
*/
#ifndef N_ROOT_H
#include "kernel/nroot.h"
#endif
#ifndef E_INCLUDES_H
#include "enkiduo/nincludes.h"
#endif
#include "enkiduo/nthing.h"

#undef N_DEFINES
#define N_DEFINES nVar
#include "kernel/ndefdllclass.h"

#define CHECK_SETS
class nEnt;
//------------------------------------------------------------------------------
const float FLOAT_TOLERANCE = .00001f;
const float QUAT_TOLERANCE = .0000001f;
class var;
class nThing;
class nVarServer;
enum RC_DEFAULT_TYPES {
    RC_LINE,
};
enum VarType { 
    VARTYPE_VOID,
    VARTYPE_INT,
    VARTYPE_FLOAT,
    VARTYPE_VECTOR3,
    VARTYPE_QUATERNION,
    VARTYPE_STRING,
    VARTYPE_REF,
    VARTYPE_RC,
    VARTYPE_OBJECT,
};

enum VARCAT{
    VARCAT_VOID,
    VARCAT_CLIENT,
    VARCAT_SERVER,
};
class nVar : public nRoot
{
public:
    nVar():
        type(VARTYPE_VOID),
        counter(0),
        rc_ptr(0),
        childState(-1),
        category(VARCAT_VOID),
        ent(0),
        claimed( false )
    {
        limits.min = -100000000000.0f;
        limits.max =  100000000000.0f;
        if( varServer == 0 )
            varServer = (nVarServer*) kernelServer->Lookup("/sys/servers/var");
    }
    virtual ~nVar(){
        ClearMem();
    }
    virtual bool SaveCmds(nPersistServer* persistServer);
    inline VarType GetType(){ return type; }
    void InheritFrom( nVar* v );
    void ToString( char* s );
    void AddChildVar( var* childVar, const char* varName );
    void RemoveChildVar( var* childVar );
    void UpdateChildren();
    bool isvalid();

    ////////////////////////////////ACCESS FUNCTIONS////////////////////////////////////////
    inline void SetI(const int i);
    inline void SetF(const float f);
    inline void SetV(const vector3& v);
    inline void SetV(const float x, const float y, const float z);
    inline void SetQ(const quaternion& q);
    inline void SetQ(const float x, const float y, const float z, const float w);
    inline void SetS(const nString& s);
    inline void SetS(const char* s);
    inline void SetO(void* const object );
    inline void SetRef( const nString& s );
    void SetRef( nEnt* t );
    inline int          GetI();
    inline float        GetF();
    inline vector3&     GetV();
    inline quaternion&  GetQ();
    inline nString&     GetS();
    inline nEnt*      GetRef();
    inline const char*  GetRefPath();
    inline rc*          GetRC();
    inline void         DefaultI( int i );
    inline void         DefaultF( float f );
    inline void         DefaultV( const vector3& v );
    inline void         DefaultV( float x, float y, float z );
    inline void         DefaultQ( quaternion q );
    inline void         DefaultQ( float x, float y, float z, float w );
    inline void         DefaultS( const char* s ) ;
    inline void         DefaultRef( const char* s );
    inline void         DefaultRC( RC_DEFAULT_TYPES type, float min, float max );
    inline void         DefaultInputRange( float min, float max, int samples );
    inline void         DefaultOutputRange( float min, float max );
    inline void         LoadRC( const char* filename );
    inline void         SaveRC( const char* filename );

    inline float        ToYaw();
    inline int          GetCounter(){ return counter; }

    /* RC wrapper funcs */
    inline float        Response( float f );
    inline void         SetInputRange( float newMin, float newMax, int samples );
    inline void         SetOutputRange( float outMin, float outMax );
    inline void         SetBucketEdge(   int index, float val );

    inline int          GetBucketCount();
    inline float        GetBucketEdge( int i );
    inline void         GetInputRange( float& min, float&max, int& samples );
    inline void         GetOutputRange( float& min, float& max );

    inline float         GetLastChecked();

    virtual void 	SetCategory( VARCAT category );
    virtual VARCAT	GetCategory(){ return category; }

    static nKernelServer* kernelServer;
    inline void         SetRange( float min, float max ) {
        limits.min = min;
        limits.max = max;
    }
    inline void       GetRange( float& min, float& max ) const{
        min = limits.min;
        max = limits.max;
    }
    virtual bool Release();
    virtual int  ReleasePriority(); //release after aspects
    virtual void FindEnt();
    virtual void SetClaimed( bool b )   { this->claimed = b; }
    virtual bool GetClaimed()           { return claimed; }
    void DontCount()                    { this->counter--; }
protected:
    VARCAT category;
    void SetType( VarType varType );
    void CheckType( VarType varType );
    void ClearMem();
    VarType type;
    int counter;
    bool claimed;
    static nVarServer* varServer;
    struct ChildVarInfo{
        var* childVar;
        nString   childName;
    };
    union{
        int                     intVal;
        float                   floatVal;
        void*                   objectVal;

        vector3*                vectorVal;
        quaternion*             quatVal;
        nString*                stringVal;
        struct {
            nDynAutoRef<nEnt>*    ref;
            nArray<ChildVarInfo*>*  childVars;
            int childState;
        };
    };
    rc*                     	rc_ptr;
    struct {
        float min, max;
    } limits;
    nEnt* ent;
};
//------------------------------------------------------------------------------

inline void nVar::SetI(const int i){
    SetType(VARTYPE_INT);
#ifdef CHECK_SETS
    if( intVal != i ) {
#endif
        intVal = i;
        counter++;
#ifdef CHECK_SETS
    }
#endif
}
inline void nVar::SetF(const float f){
    SetType(VARTYPE_FLOAT);
#ifdef CHECK_SETS
    if( n_fequal( floatVal, f, FLOAT_TOLERANCE ) == false ) {
#endif
        floatVal = f;
        counter++;
#ifdef CHECK_SETS
    }
#endif
}
inline void nVar::SetV(const vector3& v){
    SetType(VARTYPE_VECTOR3);
#ifdef CHECK_SETS
    if( vectorVal == 0 || vectorVal->isequal( v, FLOAT_TOLERANCE ) == false ) {
#endif
        *vectorVal = v;
        counter++;
#ifdef CHECK_SETS
    }
#endif
}
inline void nVar::SetV(const float x, const float y, const float z){
    SetType(VARTYPE_VECTOR3);
#ifdef CHECK_SETS
    if( vectorVal == 0 || vectorVal->isequal( vector3(x,y,z), FLOAT_TOLERANCE ) == false ) {
#endif
        vectorVal->set(x,y,z);
        counter++;
#ifdef CHECK_SETS
    }
#endif
}
inline void nVar::SetQ(const quaternion& q){
    SetType(VARTYPE_QUATERNION);
#ifdef CHECK_SETS
    if( quatVal == 0 || quatVal->isequal( q, QUAT_TOLERANCE ) == false) {
#endif
        *quatVal = q;
        counter++;
#ifdef CHECK_SETS
    }
#endif
}
inline void nVar::SetQ(const float x, const float y, const float z, const float w){
    SetType(VARTYPE_QUATERNION);
#ifdef CHECK_SETS
    if( quatVal == 0 || quatVal->isequal( quaternion(x,y,z,w), QUAT_TOLERANCE ) == false ) {
#endif
        quatVal->set(x,y,z,w);
        counter++;
#ifdef CHECK_SETS
    }
#endif
}
inline void nVar::SetS(const nString& s){
    SetType(VARTYPE_STRING);
#ifdef CHECK_SETS
    if( stringVal == 0 || *stringVal != s ) {
#endif
        *stringVal = s;
        counter++;
#ifdef CHECK_SETS
    }
#endif
}
inline void nVar::SetS(const char* s){
    SetType(VARTYPE_STRING);
#ifdef CHECK_SETS
    if( stringVal == 0 || *stringVal != s ) {
#endif
        *stringVal = s;
        counter++;
#ifdef CHECK_SETS
    }
#endif
}
inline void nVar::SetO(void* const object ){
    SetType(VARTYPE_OBJECT);
    objectVal = object;
    counter++;
}
inline void nVar::SetRef( const nString& s ){
    //n_printf("%s->SetRef(%s)\n", this->GetName(), s.Get() );
    SetType(VARTYPE_REF);
    *ref = s.Get();
    counter++;
    //this->UpdateChildren();
    this->childState = -1;
}
inline int nVar::GetI(){
    CheckType(VARTYPE_INT);
    return intVal;
}
inline float nVar::GetF(){
    CheckType(VARTYPE_FLOAT);
    return floatVal;
}
inline vector3& nVar::GetV(){
    CheckType(VARTYPE_VECTOR3);
    return *vectorVal;
}
inline quaternion& nVar::GetQ(){
    CheckType(VARTYPE_QUATERNION);
    return *quatVal;
}
inline nString& nVar::GetS(){
    CheckType(VARTYPE_STRING);
    return *stringVal;
}
inline nEnt* nVar::GetRef(){
    CheckType(VARTYPE_REF);
    if( counter == 0 )
        return 0;
    if( ref->isvalid() )
        return ref->get();
    else
        return 0;
}
inline const char* nVar::GetRefPath(){
    CheckType(VARTYPE_REF);
    return ref->getname();
}
inline rc* nVar::GetRC(){ 
    CheckType(VARTYPE_RC);
    n_assert( rc_ptr );
    return rc_ptr;
}
inline void nVar::DefaultI( int i ){
    if( counter == 0 )
        SetI( i );
}
inline void nVar::DefaultF( float f ){
    if( counter == 0 )
        SetF( f );
}
inline void nVar::DefaultV( const vector3& v ){
    if( counter == 0 )
        SetV( v );
}
inline void nVar::DefaultV( float x, float y, float z ){
    if( counter == 0 )
        SetV( x,y,z );
}
inline void nVar::DefaultQ( quaternion q ){
    if( counter == 0 )
        SetQ( q );
}
inline void nVar::DefaultQ( float x, float y, float z, float w ){
    if( counter == 0 )
        SetQ( x,y,z,w );
}
inline void nVar::DefaultS( const char* s ) {
    if( counter == 0 )
        SetS( s );
}
inline void nVar::DefaultRef( const char* s ){
    if( counter == 0 )
        SetRef( s );
}
//RC wrappers
//Sets the RC to be a simple math curve
//possibilties are currently = straight line.
inline void nVar::DefaultRC( RC_DEFAULT_TYPES type, float min, float max ){
    if( counter == 0 ) {
        SetType(VARTYPE_RC);
        switch( type ){
            case( RC_LINE ):
	int s = GetBucketCount();
	for( int i = 0; i < s; i ++ )
	    SetBucketEdge( i, min + (max-min)*(float)i/(float)(s-1));
	break;
        }
        counter++;
    }
}
inline float nVar::Response( float f ){
    n_assert( rc_ptr );
    return rc_ptr->Response( f );
}
    inline void nVar::DefaultInputRange( float min, float max, int samples ){
        if( counter == 0 )
            this->SetInputRange( min, max, samples );
    }
inline void nVar::SetInputRange( float newMin, float newMax, int samples ){
    SetType( VARTYPE_RC );
    n_assert( rc_ptr );
    rc_ptr->SetInputRange( newMin, newMax, samples );
}
    inline void nVar::DefaultOutputRange( float min, float max ){
        if( counter == 0 )
            this->SetOutputRange( min, max );
    }
inline void nVar::SetOutputRange( float outMin, float outMax ){
    SetType( VARTYPE_RC );
    n_assert( rc_ptr );
    rc_ptr->SetOutputRange( outMin, outMax );
}
inline void nVar::SetBucketEdge(   int index, float val ){
    n_assert( rc_ptr );
    rc_ptr->SetBucketEdge( index, val );
    counter++;
}
inline int nVar::GetBucketCount(){
    n_assert( rc_ptr );
    return rc_ptr->GetBucketCount();
}
inline float nVar::GetBucketEdge( int i ){
    n_assert( rc_ptr );
    return rc_ptr->GetBucketEdge( i );
}
inline void nVar::GetInputRange( float& min, float&max, int& samples ){
    n_assert( rc_ptr );
    rc_ptr->GetInputRange( min, max, samples );
}
inline void nVar::GetOutputRange( float& min, float& max ){
    n_assert( rc_ptr );
    rc_ptr->GetOutputRange( min, max );
}
inline void nVar::LoadRC( const char* filename ) {
    SetType(VARTYPE_RC);
    n_assert( rc_ptr );
    rc_ptr->Load( filename );
    this->counter++;
}
inline void nVar::SaveRC( const char* filename ) {
    CheckType(VARTYPE_RC);
    n_assert( rc_ptr );
    rc_ptr->Save( filename );
}
inline float nVar::GetLastChecked() {
    CheckType(VARTYPE_RC);
    n_assert( rc_ptr );
    return rc_ptr->GetLastChecked();
}
#endif
