import math
from utils_navigation import *
from cat import behaviorInfo



def directCPA(p0x, p0y, up, bp, q0x, q0y, vq, bq):
    # Computes the time and distance CPA
    # 

    # Input:
    # p0x, p0y - initial coordinates of the reference ship (in meters)
    # up - speed of the reference ship (in knots)
    # bp - heading of the reference ship (in degrees)
    # q0x, q0y - initial coordinates of own ship (in meters)
    # vq - speed of own ship (in knots)
    # bq - heading of own ship (in degrees)

    # Returns:
    # t - time to CPA (in seconds)
    # distCPA - distance at CPA (in meters)

    # Convert angles to radians
    bp_rad = math.radians(bp)
    bq_rad = math.radians(bq)


    w0x = p0x - q0x
    w0y = p0y - q0y

    w0_squared = w0x**2 + w0y**2
    #print "w0x = ", w0x, "w0y = ", w0y, "w0_squared = ", w0_squared
    
    ux = up*math.cos(bp_rad)
    uy = up*math.sin(bp_rad)
    if (ux > 0 and ux < 1.0e-5):
        ux = 0
    if (uy > 0 and uy < 1.0e-5):
        uy = 0
    #print "ux = ", ux, "uy = ", uy
    

    vx = vq*math.cos(bq_rad)
    vy = vq*math.sin(bq_rad)
    if vx > 0 and vx < 1.0e-5:
        vx = 0
    if vy > 0 and vy < 1.0e-5:
        vy = 0

    u_min_v_squared = (ux-vx)**2 + (uy-vy)**2
    if u_min_v_squared == 0:
        return 0, math.sqrt(w0_squared)

    w0x_uv = (w0x * (ux - vx) + w0y * (uy - vy))

    #print " u_min_v_squared = ", u_min_v_squared, " w0x_uv = ", w0x_uv, " w0_squared = ", w0_squared
    
    t = - w0x_uv/u_min_v_squared
    dist_CPA =  u_min_v_squared * t**2 + 2 * w0x_uv * t + w0_squared

    if  dist_CPA < 1.0e-5:
        dist_CPA = 0
        
    if dist_CPA < 0:
        return 0, t
    
    dist_CPA = math.sqrt(dist_CPA)
    dist_CPA = dist_CPA
    #print "tCPA = ", t, "distCPA = ", dist_CPA
    return (t, dist_CPA)


def inverseCPA(p0x, p0y, up, bp, q0x, q0y, vq, bq, cpa_dist, s):
    # Computes the speed and angle of a boat that would increase the CPA
    # to cpa_dist

    # Input:
    # p0x, p0y - initial coordinates of the reference ship (yards)
    # up - speed of the reference ship (in meters per second)
    # bp - heading of the reference ship (in degrees)
    # q0x, q0y - initial coordinates of own ship (yards)
    # vq - speed of own ship (in meters per second)
    # bq - heading of own ship (in degrees)
    # s - min speed desired for the maneuver (in meters per second)

    # Returns:
    # existSolution - boolean, describing if a solution exists
    # v - speed at which maneuver should be performed
    # b - course at which maneuver should be performed


    # Convert from trig. coordinates to this function's coordinate system
##     bp = NormalizeHeading((360 - bp) + 90)
##     bq = NormalizeHeading((360 - bq) + 90)

    # Convert angles to radians
    bp_rad = math.radians(bp)
    bq_rad = math.radians(bq)

    w0x = p0x - q0x
    w0y = p0y - q0y
    w0_squared = w0x**2 + w0y**2

    a = cpa_dist**2 - w0_squared
    b = a + w0x**2
    c = a + w0y**2
    d = 2*w0x*w0y

    ux = up*math.cos(bp_rad)
    uy = up*math.sin(bp_rad)

    vx = math.cos(bq_rad)
    vy = math.sin(bq_rad)
    
    e = b * (ux**2) + c * (uy**2) + d * ux * uy
    m = b * vx**2 + c * vy**2 + d * vx * vy
    n = - (2*b*ux*vx + 2*c*uy*vy + d*ux*vy + vx*uy)

    delta = n**2 - 4*m*e

    if delta < 0:
        #print "Case 2"
        return (0, 0, 0)
        # Compute a new heading for which delta = 0
##         q = 4*(b**2)*(ux**2) + 4*b*ux*uy + uy**2 - 4*b*e
##         p = 4*(c**2)*(uy**2) + 4*c*d*ux*uy + (d**2)*(ux**2) - 4*c*e
##         r = 8*b*c*ux*uy + 4*b*d*(ux**2) + 4*c*(uy**2) + 2*d*ux*uy-4*d*e

##         delta1 = r**2 - 4*q*p
##         if delta1<0:
##             print "*************>>>>>>>>No solution to the problem exists"
##             return (0, 0, 0)
##         else:
##             t1 = math.atan2 (-r + math.sqrt(delta1), 2 * p)
##             t2 = math.atan2 (-r - math.sqrt(delta1), 2 * p)
##             b = t1*180/math.pi
##             #b = NormalizeHeading((360 - b) + 90)
##             b2 = t2*180/math.pi
##             #b2 = NormalizeHeading((360 - b2) + 90)

##             print "n = ", n, "m = ", m, "e = ", e
##             v = -n/(2*m)
## ##             if v < 5:
## ##                 v = 5
##             if v < 0:
##                 v = s
##             print "InverseCPA2: heading = ", b, " heading2 = ", b2, " speed = ", v
##             return (1, v, b)

    #print "Case 1"
    v1 = (-n + math.sqrt(delta))/(2*m)
    v2 = (-n - math.sqrt(delta))/(2*m)

    if v1 <= 0 and v2 <= 0:
        # return a bearing for a speed of zero
        delta1 = d**2 - 4 *c*b
        if delta1 <=0:
##             b = NormalizeHeading((360 - bq) + 90)
##             v = 5
##             print "Inverse CPA1: v = ", 0, "b = ", b
##             return (v, b)
            #print "*************>>>>>>>>No solution to the problem exists"
            return (0, 0, 0)
        
        #print "delta1 = ", delta1
        t1 = math.atan2 (-d+math.sqrt(delta1), 2*c)
        t2 = math.atan2 (-d-math.sqrt(delta1), 2*c)

        #print t1*180/math.pi, t2*180/math.pi
        b = t1*180/math.pi
        #b = NormalizeHeading((360 - b) + 90)
        b2 = t2*180/math.pi
        #b2 = NormalizeHeading((360 - b2) + 90)
        v = 0
        #print "Inverse CPA3: v = ", v, "b = ", b, "b2 = ", b2
        return (1, 0, b)
    else:
        if v1 > 0:
            #v = (v1 * 60)/2000
            sol = 1
            v = v1
            b = bq
            #b =  NormalizeHeading((360 - bq) + 90)
        elif v2 > 0:
            sol = 1
            #v = (v1 * 60)/2000
            v = v2
            b = bq
            #b =  NormalizeHeading((360 - bq) + 90)
        else:
            print "Error: should not get here"
            sol = 0
            v = 0
            b = 0
    #print "Inverse CPA4: v = ", v, "b = ", b
    return (sol, v, b)
