#!/usr/bin/env python
"""
    Spiffy XML Parser
    Reads in two behavior networks from xml files
    Merges them into a resultant network
    Saves the merged network out to a new file

    Uses dom parsing for all the xml
    Works inside the xml tree - should maintain any additional information it doesn't know about
"""
__author__ = 'Chris Miles (miles@cse.unr.edu) and Adam Olenderski (olenders@cse.unr.edu)'
__version__ = '1.1'
parameterEqualityThreshold = 10     #the tolerance thereshold for parameter differences (note the sum of differences for all params must be < this)

import sys
import inspect
import math
import string
from xml.dom import minidom
from xml.dom.minidom import getDOMImplementation
from xml.dom.minidom import parseString
#-----------------------------------------------------------------------------------------------------------------------------
def cleanup( root, doc ):
    rc = ""
    for node in root.childNodes:
	if node.nodeType == node.TEXT_NODE:
	    rc = rc + node.data
	    node.unlink()
	    root.removeChild( node )
    for node in root.childNodes:
    	cleanup( node,doc )

    rc = rc.strip()
    if rc:
	textNode = doc.createTextNode(rc)
	root.appendChild( textNode )

class RowTree:
    """
    Contains the tree of rows used to calculate the LCS for a DAG
    members: 2 lists of the row entries, a reference to the parent
    row, a list of child rows, an x index, and a list of Y indices
    """
    def __init__ (self):
        """initialize to an empty row"""
        self.c = []
        self.b = []
        self.parentRow = 0
        self.children = []
        self.X = 0
        self.Y = []
        self.bestRow = 0
    def __str__(self):
        """cast me to a string"""
        st = str(self.c) + ',' + str(self.b)+'\n'
        for child in self.children:
            st+=str(child)
        return st
    def __repr__(self):
        return str(self)
    def addRow(self,X_,Y_,currentParent):
        #print "Adding a Row"
        if currentParent == 0:
            self.X = X_
            self.Y = Y_
            for i in range(len(Y_)):
                #initialize the matrix's rows
                self.b += [0]
                self.c += [0]
            for i in range(len(Y_)):
                if self.Y[i] == self.X:
                    self.b[i] = 'ul'
                    self.c[i] = 1
                elif i == 0:
                    self.b[0] = 'up'
                    self.c[0] = 0
                elif self.c[i-1] == 0:
                    self.b[i] = 'up'
                    self.c[i] = 0
                else:
                    self.b[i] = 'left'
                    self.c[i] = 1
            #print self.b,self.c
            return self
        
        else:
            newRow = RowTree()
            newRow.parentRow = currentParent
            newRow.X = X_
            newRow.Y = Y_
            for i in range(len(newRow.Y)):
                #initialize the matrix's rows
                newRow.b += [0]
                newRow.c += [0]
            for item in newRow.Y:
                i = newRow.Y.index(item)
                if item == newRow.X:
                    newRow.b[i] = 'ul'
                    if i == 0:
                        newRow.c[i] = 1
                    else:
                        newRow.c[i] = newRow.parentRow.c[i-1] + 1
                elif i == 0:
                    newRow.b[i] = 'up'
                    newRow.c[i] = newRow.parentRow.c[i]
                elif newRow.c[i-1] > newRow.parentRow.c[i]:
                    newRow.c[i] = newRow.c[i-1]
                    newRow.b[i] = 'left'
                else:
                    newRow.c[i] = newRow.parentRow.c[i]
                    newRow.b[i] = 'up'
            newRow.parentRow.children += [newRow] #might need to change this to newRow.parentRow.children = newRow.parentRow.children + newRow
        #print newRow.b,newRow.c
        return newRow

class Trigger:
    def __init__(self):
        self.name = "unnamed trigger"
        self.id = -1
        self.xmlTree = 0
        self.params = []
        self.slots = []
    #def __str__(self):
    def addParam(self, name,type,value):
        if type=='bool':
            self.params.append({'name':name, 'type':type, 'value':bool(value)})
        elif type=='int':
            self.params.append({'name':name, 'type':type, 'value':int(value)})
        elif type=='float':
            self.params.append({'name':name, 'type':type, 'value':float(value)})
        else:
            self.params.append({'name':name, 'type':type, 'value':value})
    def addSlot(self,name,id):
        self.slots.append({'name':str(name),'id':str(id)})
    def updateXml(self,document):
        self.xmlTree.setAttribute('name',str(self.name))
        self.xmlTree.setAttribute('id',str(self.id))
        for p in self.xmlTree.getElementsByTagName('parameter'):
            self.xmlTree.removeChild(p)
            p.unlink()
        for param in self.params:
            node = document.createElement('parameter')
            node.setAttribute('name',str(param['name']))
            node.setAttribute('type',str(param['type']))
            node.setAttribute('value',str(param['value']))
            self.xmlTree.appendChild(node)
        for s in self.xmlTree.getElementsByTagName('slot'):
            self.xmlTree.removeChild(s)
            s.unlink()
        for slot in self.slots:
            node = document.createElement('slot')
            node.setAttribute('id',str(slot['id']))
            node.setAttribute('name',str(slot['name']))
            self.xmlTree.appendChild(node)
            
        






class Slot:
    def __init__(self):
        self.name = 'unnamed slot'
        self.id = -1
        self.xmlTree = 0
    def updateXml(self,document):
        self.xmlTree.setAttribute('name',str(self.name))
        self.xmlTree.setAttribute('id',str(self.id))
    def __str__(self):
        s = '('+str(self.name)+','+str(self.id)+')'
        return s
    def __repr__(self):
        return str(self)
    
            
        

class Behavior:
    """
    Contains a behavior,
    members: parameters, sensors, links
    """
    def __init__ (self):
        """initialize to an empty behavior"""
        self.xmlTree = 0                    #the root of my xmlTree
        self.name = "unnamed behavior"      # my name
        self.id = -1                        #my unique(within my network) identifier
        self.group = -1
        self.params = []                    #set of my parameters
        self.sensors = []                   #set of my sensors
        self.links = []                     #set of my links
    def __str__(self):  
        """cast me to a string"""
        st = str(self.id) + " ("
        for i, link in enumerate( self.links ):
            if i > 0:
                st += ','
            st += str(link['from']) + "->" + str(link['to'])
        st += ")"
        return st
    def __repr__(self):                                 
        """link to __str__"""
        return str(self)
    def __eq__(self, other ):
        """checks if two behaviors are equivalent, for now matchs parameters / sensors"""
        if self.name != other.name:
            return False
        for s in self.sensors + other.sensors:
            s['matches'] = False
        for s1 in self.sensors:
            for s2 in other.sensors:
                if s1['type'] == s2['type'] and s1['name'] == s2['name']:
                    #print "matching sensors:", s1, s2
                    s1['matches'] = True
                    s2['matches'] = True
        for s in self.sensors + other.sensors:
            if s['type'] == 'pre' and not s['matches']:
                #print "failed to match sensor", s, "returning false"
                return False
            #else:
                #print 'matched sensor',s
        #check parameters
        totalError = 0
        for p1,p2 in zip(self.params, other.params): #check if parameters match
            if p1['type'] == p2['type'] and p1['type'] == 'bool':
                if p1['value'] != p2['value']:
                    return False
            elif p1['type'] == p2['type'] and p1['type'] == 'vector':
                sumofsquares = 0
                for c1,c2 in zip(p1['value'],p2['value']):
                    sumofsquares += (c1-c2)**2
                if math.sqrt(sumofsquares) > 10.0:
                    return False                    
            else:
                totalError += float(abs(p1['value'] - p2['value']))
        return totalError < parameterEqualityThreshold
    def addParam    (self, name, min,max,value,type):            
        """append a parameter, no check for duplicates"""
        if(type == 'float'):
            self.params.append(     {'name':name, 'min':float(min), 'max':float(max), 'value':float(value), 'type':'float'           } )
        elif(type == 'bool'):
            self.params.append(     {'name':name, 'min':False, 'max':True, 'value':bool(value),'type':'bool' })
        elif(type == 'int'):
            self.params.append(     {'name':name, 'min':int(min), 'max':int(max), 'value':int(value) , 'type':'int'})
        elif(type == 'vector'):
            v = ()
            for coord in value[1:-1].split(','):
                v = v + (float(coord),)
            self.params.append(     {'name':name, 'min':float(min), 'max':float(max), 'value':v, 'type':'vector'})
    def addSensor   (self, name, type ):   
        """append a sensor, no check for duplicates"""
        #print 'adding sensor: name=',name,'type=',type
        self.sensors.append(    {'name':int(name),           'type':type} )
    def addLink     (self, frm, to, type ):       
        """add a new link to this behavior, no duplicates"""
        for link in self.links:
            assert link['from'] == self.id
            if link['to'] == to and link['from'] == frm:    #matches the one we are trying to add
                #print "ignoring duplicate link", link
                return              #just return for now - take the earliest link
        self.links.append(      {'from':int(frm),       'to':int(to),           'type':type         } )
    def hasLinkTo (self, other):
        """return True if there is a link from this behavior to the other (based on id)"""
        for link in self.links:
            assert link['from'] == self.id #must come from me            
            if link['to'] == other.id:
                return True
        return False
    def rename ( self, newId ):
        """changes my id - updating any of my links"""
        #print self.id, "being renamed too", newId
        for link in self.links:
            assert link['from'] == self.id
            link['from'] = newId
        self.id = newId
    def updateXml (self, document):
        """updates my xml tree"""
        #print self, ".update xml"
        self.xmlTree.setAttribute('name',   str(self.name))
        self.xmlTree.setAttribute('id',     str(self.id))
        self.updateLinkXML  (document)
        self.updateSensorXML(document)
    def updateLinkXML (self, document):
        """update our xml tree to reflect link changes"""
        #take out existing links
        for pr in self.xmlTree.getElementsByTagName('links'):
            for p in pr.getElementsByTagName('to'):
                pr.removeChild( p )
                p.unlink()
            for p in pr.getElementsByTagName('from'):
                pr.removeChild(p)
                p.unlink()
            for txt in pr.childNodes:
                pr.removeChild( txt )
        #put our new links in our first links node
        linksNodes = self.xmlTree.getElementsByTagName('links')
        if len(linksNodes) == 0:
            links = document.createElements('links')
            xmlTree.appendChild( links )
        else:
            links = linksNodes[0]
        #add in our links
        for link in self.links:
            #print "adding link:", link,"to xml"
            node = document.createElement('to')
            #node.setAttribute('from',   str(link['from']))
            #node.setAttribute('to',     str(link['to']))
            node.setAttribute('type',   str(link['type']))
            textNode = document.createTextNode(str(link['to']))
            node.appendChild(textNode)
            links.appendChild( node )
            #links.appendChild( textNode )
    def updateSensorXML (self, document):
        """update our xml tree to reflect sensor changes"""
        #take out existing links
        for pr in self.xmlTree.getElementsByTagName('sensors'):
            for p in pr.getElementsByTagName('pre'):
                pr.removeChild( p )
                p.unlink()
            for p in pr.getElementsByTagName('post'):
                pr.removeChild(p)
                p.unlink()
            for txt in pr.childNodes:
                pr.removeChild( txt )
        #put our new sensors in
        sensorsNodes = self.xmlTree.getElementsByTagName('sensors')
        if len(sensorsNodes) == 0:
            sensors = document.createElements('sensors')
            xmlTree.appendChild( sensors )
        else:
            sensors = sensorsNodes[0]
        for sensor in self.sensors:
            if sensor['type'] == 'pre':
                node = document.createElement('pre')
            else:
                node = document.createElement('post')
                
            node.setAttribute('id',     str(sensor['name']))
            sensors.appendChild( node )
            #sensors.appendChild( textNode )
#-----------------------------------------------------------------------------------------------------------------------------
class BehaviorNetwork:
    """
    Contains a network of behaviors
    functions for loading/saving to xml, and merging with other networks
    """
    def __init__(self):
        """initializes with an empty list of behaviors"""
        self.name = 'unnamed'           #my name
        self.behaviors = []             #list of my behaviors
        self.triggers = []              #list of my triggers
        self.slots = []                 #list of my slots
    def __str__ (self):
        """return my behaviors as a string"""
        return str( self.behaviors )
    def __repr__ (self):    
        """link to __str__"""
        return str(self)
    def loadXMLFile( self, filename, xmlstring=False ):
        """loads this network from an xml file"""
        if xmlstring:
            self.document = parseString(filename)
        else:
            file = open( filename )
            self.document = minidom.parse( file )
            file.close()
        self.xmlRoot = self.document.documentElement

        for t in self.xmlRoot.getElementsByTagName('triggers'):  #get each trigger
            for tr in t.getElementsByTagName('trigger'):
                trigger = self.addTrigger()
                trigger.xmlTree = tr.cloneNode(True)
                trigger.name = tr.getAttribute('name')
                trigger.id = int(tr.getAttribute('id'))
                for p in tr.getElementsByTagName('parameter'):
                    trigger.addParam( p.getAttribute('name'),p.getAttribute('type'),p.getAttribute('value'))
                for s in tr.getElementsByTagName('slot'):
                    trigger.addSlot(str(s.getAttribute('name')),str(s.getAttribute('id')))

        for s in self.xmlRoot.getElementsByTagName('slots'):   #get each slot
            for sl in s.getElementsByTagName('entity'):
                slot = self.addSlot()
                slot.name = sl.getAttribute('name')
                slot.id = sl.getAttribute('id')
                slot.xmlTree = sl.cloneNode(True)
        
        for e in self.xmlRoot.getElementsByTagName('behavior'): #foreach behavior
            behavior = self.addBehavior()
            behavior.xmlTree    = e.cloneNode(True)             #copy that piece of xml into the behavior
            behavior.name       = e.getAttribute('name')        #set the name
            behavior.id         = int(e.getAttribute('id'))     #set the id
            behavior.group      = int(e.getAttribute('group'))
            for pr in e.getElementsByTagName('parameters'):
				#The individual parameters now have the following values: name, min, max. The tag name is
				#'float', 'bool', or 'int'. Containes in the tags is the default values:
				#<float name="speed" min="0.0" max="22.22">18.0</float>
                for p in pr.getElementsByTagName('float'):
                    behavior.addParam(  p.getAttribute('name'),p.getAttribute('min'),p.getAttribute('max'),p.childNodes[0].data,'float')
                for p in pr.getElementsByTagName('bool'):
                    behavior.addParam(  p.getAttribute('name'),False,True,p.childNodes[0].data,'bool')
                for p in pr.getElementsByTagName('int'):
                    behavior.addParam(  p.getAttribute('name'),p.getAttribute('min'),p.getAttribute('max'),p.childNodes[0].data,'int')
                for p in pr.getElementsByTagName('vector'):
                    behavior.addParam(  p.getAttribute('name'),p.getAttribute('min'),p.getAttribute('max'),p.childNodes[0].data,'vector')
                    
                   # behavior.addParam(  p.getAttribute('id'), p.childNodes[0].data )

            for sr in e.getElementsByTagName('sensors'):
                for s in sr.getElementsByTagName('post'):                                       
                    behavior.addSensor( s.getAttribute('id'),'post')
                for s in sr.getElementsByTagName('pre'):
                    behavior.addSensor( s.getAttribute('id'),'pre')

            for lr in e.getElementsByTagName('links'):
				#links now can have two tags: <to> and <from>. They each have an attribute of 'type' and inside
				#the open and close tags are the destination or origin. Right now, I only use <to> tags.
				#ex. <to type="ordering">4</to>
                for l in lr.getElementsByTagName('to'):
                    behavior.addLink(   behavior.id, l.childNodes[0].data, l.getAttribute('type') )
                for l in lr.getElementsByTagName('from'):
                    behavior.addLink(   l.childNodes[0].data, behavior.id, l.getAttribute('type') )
        for e in self.behaviors:
            for link in e.links:
                if link['from'] != e.id:
                    orig = link['from']
                    dest = link['to']
                    type = link['type']
                    e.links.remove(link)
                    for f in self.behaviors:
                        if f.id == orig:
                            f.addLink(orig,dest,type)
            
                    
            #self.behaviors.append( behavior )
    def toxml (self):
        """
        build my xml tree
        create a new document
        put in each behavior
        """
        document = minidom.Document()
        xmlRoot = document.createElement('network')
        #xmlTree = document.appendChild( xmlRoot )
        triggersRoot = document.createElement('triggers')
        slotsRoot = document.createElement('slots')
        
        for t in self.triggers:
            t.updateXml(document)
            triggersRoot.appendChild(t.xmlTree)
        for s in self.slots:
            s.updateXml(document)
            slotsRoot.appendChild(s.xmlTree)
        cleanup(triggersRoot,document)
        cleanup(slotsRoot,document)
        xmlRoot.appendChild(triggersRoot)
        xmlRoot.appendChild(slotsRoot)
        for b in self.behaviors:
            #print "appending behavior", b
            b.updateXml(document)
            xmlRoot.appendChild( b.xmlTree )
	cleanup( xmlRoot, document )
        return xmlRoot.toprettyxml("  ")
    def topologicalSort (self):
        """
        sorts this behavior network topologically
        i.e. behaviors appear before their children
        note. topologically sorted lists are not unique
        """
        origLength = len(self.behaviors)
        sortedBehaviors = []
        didSomething = 1
        while didSomething:                                 #while we can pull someone out of the array
            didSomething = 0
            for behavior in self.behaviors:                 #check each behavior
                found = 0
                for behavior2 in self.behaviors:            #see if anyone else has a link to me
                    if behavior != behavior2:
                        if behavior2.hasLinkTo(behavior):
                            found = True
                            break
                if not found:                               #if no one has a link to me then remove me from the list - redo the loop
                    sortedBehaviors.append( behavior )
                    self.behaviors.remove( behavior )
                    didSomething = 1
                    break
        assert len(self.behaviors) == 0
        assert len(sortedBehaviors) == origLength
        self.behaviors = sortedBehaviors
    def rename(self, fromId, toId ):
        """We have to rename one of our nodes, changes its id + any links in my network"""
        #print "networking renaming ids from", fromId, "to", toId
        numFound = 0
        for b in self.behaviors:
            if b.id == fromId:
                #print "renaming", b, "to", toId
                b.rename( toId )
                numFound += 1
                assert numFound == 1 #can have at most one behavior with any given id
            for link in b.links:
                if link['to'] == fromId:
                    #print "changing link", link, "to point to", toId
                    link['to'] = toId
    def renameTrigger(self, fromID, toID):
        numFound = 0
        #print 'renaming trigger',fromID,'to',toID
        for t in self.triggers:
            if t.id == fromID:
                t.id = toID
                numFound += 1
                assert numFound == 1
            for b in self.behaviors:
                for sensor in b.sensors:
                    #print sensor['name']
                    if sensor['name'] == fromID:
                        #print 'found a sensor using id',fromID
                        sensor['name'] = toID
                        
        assert numFound == 1

    def addTrigger(self):
        """Create and return a new trigger"""
        trigger = Trigger()
        self.triggers.append(trigger)
        return trigger
    def addSlot(self):
        slot = Slot()
        self.slots.append(slot)
        return slot
    def addBehavior (self):
        """Create and return a new behavior"""
        behavior = Behavior()                               #create the behavior
        self.behaviors.append( behavior )
        return behavior
    def adjacencyList(self,behavior):
        """returns a list of all behaviors to which this behavior links"""
        alist = []
        alistbyId = []
        for link in behavior.links:
            alistbyId += [link['to']]
        for idnum in alistbyId:
            for b in self.behaviors:
                if idnum == b.id:
                    alist += [b]
        return alist
    def isRootNode(self,behavior):
        """returns true if behavior is a rootnode in this network,false otherwise"""
        for b in self.behaviors:
            for bl in b.links:
                if bl['to'] == behavior.id:
                    return False
        return True


    def doesPrecede(self,preB,sucB):
        alist = self.adjacencyList(preB)
        if len(alist) == 0:
            return False
        else:
            for behavior in alist:
                if behavior.id == sucB.id:
                    return True
            foundSuccessor = False
            for behavior in alist:
                if self.doesPrecede(behavior,sucB):
                    foundSuccessor = True
        return foundSuccessor
                
        

#-----------------------------------------------------------------------------------------------------------------------------
def findUnusedId (networks):
    """searches a list of networks to find an unused id"""
    id = 0
    found = True
    while found:
        found = False
        for network in networks:
            for beh in network.behaviors :
                if beh.id == id:
                    found = True
                    id += 1
    #print "found id", id
    return id
#-----------------------------------------------------------------------------------------------------------------------------
def findUnusedSlotId(networks):
    id = "A"
    found = True
    while found:
        found = False
        for network in networks:
            for slot in network.slots:
                if slot.id == id:
                    found = True
                    id = string.uppercase[string.uppercase.index(id)+1]
    return id
#-----------------------------------------------------------------------------------------------------------------------------
def findUnusedTriggerId(networks):
    id = 0
    found = True
    while found:
        found = False
        for network in networks:
            for trigger in network.triggers:                
                if trigger.id == id:
                    found = True
                    id += 1
    return id
#-----------------------------------------------------------------------------------------------------------------------------
def matchBehaviorIds (a,b):
    """goes through 2 behaviors and matches up ids"""
    #print "before renaming       ", a, b
    idlist = []
    duplicateId = False
    for network in [a.behaviors,b.behaviors]:
        for behavior in network:
            duplicateId = False
            for id in idlist:
                if behavior.id == id:
                    duplicateId = True
                    newId = findUnusedId([a,b])
                    
            if duplicateId:
                newId = findUnusedId([a,b])
                a.rename(behavior.id,newId)
                idlist.append(newId)
            else:
                idlist.append(behavior.id)
                



#-----------------------------------------------------------------------------------------------------------------------------
def getSlotList(network1,network2):
    slotlist = network1.slots
    for slot in network2.slots:
        match = False
        for slot2 in slotlist:
            if slot.id == slot2.id:
                match = True
                if slot.name == slot2.name:
                    pass
                else:
                    newSlotId = findUnusedSlotId([network1,network2])
                    for i in range(len(network2.triggers)):
                        for j in range(len(network2.triggers[i].slots)):
                            if network2.triggers[i].slots[j] == slot.id:
                                network2.triggers[i].slots[j] = newSlotId
                    slot.id = newSlotId
                    slotlist.append(slot)
            elif slot.name == slot2.name:
                match = True
                for i in range(len(network2.triggers)):
                    for j in range(len(network2.triggers[i].slots)):
                        if network2.triggers[i].slots[j] == slot.id:
                            network2.triggers[i].slots[j] = slot2.id
        if match == False:
            #print slotlist
            #print slot
            slotlist.append(slot)
    return slotlist
#-----------------------------------------------------------------------------------------------------------------------------
def getTriggerList(network1,network2):
    triggerList = network1.triggers
    
    for trigger1 in network2.triggers:
        numEqual = 0
        match = False
        for trigger2 in triggerList:
            equal = True
            if trigger1.name != trigger2.name:
                equal = False
            else:
                totalError = 0
                for p1,p2 in zip(trigger1.params,trigger2.params):
                    if p1['name'] != p2['name'] and equal:
                        equal = False
                    elif equal:
                        if p1['type'] != p2['type']:
                            equal = False
                        else:
                            if p1['type'] == 'bool':
                                if p1['value'] != p2['value']:
                                    equal = False
                            else:
                                totalError+=float(abs(p1['value'] - p2['value']))
                if totalError > parameterEqualityThreshold:
                    equal = False
                else:
                    for s1,s2 in zip(trigger1.slots,trigger2.slots):
                        if s1 != s2:
                            equal = False
            if equal:
                match = True
                numEqual+=1
                assert numEqual == 1
                network2.renameTrigger(trigger1.id,trigger2.id)
            
        if not match:
            newID = findUnusedTriggerId([network1,network2])
            network2.renameTrigger(trigger1.id,newID)
            triggerList.append(trigger1)
    return triggerList


    
    
#-----------------------------------------------------------------------------------------------------------------------------
def merge(lcsTree,network1,network2):
    mergedNetwork = BehaviorNetwork()
    
    #mergedNetwork.slots = slotlist
    #print mergedNetwork.slots

    #mergedNetwork.triggers = triggerList        
    
    x = network1.behaviors
    y = network2.behaviors
    for behavior in x + y:
        behavior.common = False
    lcsidlist = []
    typelist = []
    #greatesti = len(x)-1
    greatestj = len(y)-1
    j = greatestj
    currentRow = lcsTree.bestRow
    
    while currentRow and j >= 0:
        #print "Merging based on LCS-Tree (in while loop)",currentRow.b,currentRow.c
        if currentRow.b[j] == "ul":
            #print "Found a match:"
            #print currentRow.X
            #print currentRow.Y[j]
            newBehavior = mergedNetwork.addBehavior()
            newBehavior.name = currentRow.X.name
            newBehavior.xmlTree = currentRow.X.xmlTree.cloneNode(True)
            newBehavior.id = currentRow.X.id
            lcsidlist.append(newBehavior.id)
            network2.rename(y[j].id,newBehavior.id)
            newBehavior.params = currentRow.X.params[:]
            for link in currentRow.X.links + y[j].links:
                for newlink in newBehavior.links:
                    if link['from'] == newlink['from'] and link['to'] == newlink['to']:
                        if link['type'] == 'ordering':
                            newlink['type'] = 'ordering'
                        elif link['type'] == 'enabling' and newlink['type'] == 'permanent':
                            newlink['type'] = 'enabling'
                newBehavior.addLink( link['from'], link['to'], link['type'] )
                
            for s in currentRow.X.sensors + y[j].sensors:
                s['matches'] = False
            for as in currentRow.X.sensors:
                for bs in y[j].sensors:
                    if as['type'] == bs['type'] and as['name'] == bs['name']:
                        #print "matching sensors",as,bs
                        as['matches'] = True
                        bs['matches'] = True
            for as in currentRow.X.sensors:
                if as['matches']:
                    newBehavior.sensors.append(as)
            for s in currentRow.X.sensors + y[j].sensors:
                if s['type'] == 'post' and not s['matches']:
                    newBehavior.sensors.append(s)
            currentRow.X.common = True
            y[j].common = True
            currentRow = currentRow.parentRow
            j = j-1
        elif currentRow.b[j] == 'up':
            currentRow = currentRow.parentRow
        else:
            j = j-1

    for behavior in x + y:
        if behavior.common == False:
            newBehavior = mergedNetwork.addBehavior()
            newBehavior.name = behavior.name
            newBehavior.xmlTree = behavior.xmlTree.cloneNode(True)
            newBehavior.id = behavior.id
            newBehavior.params = behavior.params[:]
            newBehavior.sensors = behavior.sensors[:]
            newBehavior.links = behavior.links[:]

    for b1 in mergedNetwork.behaviors:
        for b2 in mergedNetwork.behaviors:
            if b1.id != b2.id:
                linkPresent = False
                for link in b1.links:
                    if link['to'] == b2.id:
                        linkPresent = True
                for link in b2.links:
                    if link['to'] == b1.id:
                        linkPresent = True
                if not linkPresent:        
                    if mergedNetwork.doesPrecede(b1,b2):
                        b1.addLink(b1.id,b2.id,'ordering')
                    elif mergedNetwork.doesPrecede(b2,b1):
                        b2.addLink(b2.id,b1.id,'ordering')
    mergedNetwork.topologicalSort()
    return mergedNetwork


#-----------------------------------------------------------------------------------------------------------------------------
def generalize(network1,network2):
    myTree = RowTree()
    myTree.bestRow = 0
    #for behavior in network1.behaviors:
    #    if network1.isRootNode(behavior):
    ExpandNode(myTree,network1.behaviors[0],network1,network2,0)
    return myTree

#-----------------------------------------------------------------------------------------------------------------------------
def ExpandNode(tree,behavior,network1,network2,parentRow):
    myRow = tree.addRow(behavior,network2.behaviors,parentRow)
    if parentRow == 0:
        tree.bestRow = myRow
    elif myRow.c[len(network2.behaviors)-1] >= tree.bestRow.c[len(network2.behaviors)-1]:
        tree.bestRow = myRow
    for b in network1.adjacencyList(behavior):
        ExpandNode(tree,b,network1,network2,myRow)
    return

#-----------------------------------------------------------------------------------------------------------------------------
def mergeNetworks(network1,network2):
    if isinstance(network1,str):
        net1 = BehaviorNetwork()
        net1.loadXMLFile(network1)
    elif isinstance(network1,BehaviorNetwork):
        net1 = network1
        
    net1.topologicalSort()

    if isinstance(network2,str):
        net2 = BehaviorNetwork()
        net2.loadXMLFile(network2)
    elif isinstance(network2,BehaviorNetwork):
        net2 = network2

    net2.topologicalSort()
    
    matchBehaviorIds(net1,net2)
    slotList = getSlotList(net1,net2)
    triggerList = getTriggerList(net1,net2)
    lcsTree = generalize(net1,net2)
    mergedNetwork = merge(lcsTree,net1,net2)
    mergedNetwork.slots = slotList
    mergedNetwork.triggers = triggerList
    return mergedNetwork
    
#-----------------------------------------------------------------------------------------------------------------------------
def mergeNetworkStrings(network1,network2):
    net1 = BehaviorNetwork()
    net1.loadXMLFile(network1,xmlstring=True)
    net1.topologicalSort()
    net2 = BehaviorNetwork()
    net2.loadXMLFile(network2,xmlstring=True)
    net2.topologicalSort()
    matchBehaviorIds(net1,net2)
    slotList = getSlotList(net1,net2)
    triggerList = getTriggerList(net1,net2)
    lcsTree = generalize(net1,net2)
    mergedNetwork = merge(lcsTree,net1,net2)
    mergedNetwork.slots = slotList
    mergedNetwork.triggers = triggerList
    return mergedNetwork.toxml()

    
        
#-----------------------------------------------------------------------------------------------------------------------------        
def main(argv): 
    """
    The main function, 
    reads in the two xml files, 
    create the two source behaviors, 
    calls the merge function, 
    and dumps out the resultant
    """
    networks = [BehaviorNetwork(), BehaviorNetwork()]

    if argv ==[]:
        networks[0].loadXMLFile('test1.xml')
        networks[1].loadXMLFile('test2.xml')
        networks[0].topologicalSort()
        networks[1].topologicalSort()
        matchBehaviorIds(networks[0],networks[1])

        lcsTree = generalize(networks[0],networks[1])
                
        mergedNetwork = merge(lcsTree,networks[0],networks[1])
        
    else:
        mergedNetwork = mergeNetworks(argv[0],argv[1])
        #networks[0].loadXMLFile(argv[0])
        #networks[0].topologicalSort()
        #sys.stderr.write(arg)
        #continue
        for arg in argv[2:]:
            mergedNetwork = mergeNetworks(mergedNetwork,argv[arg])
            #networks[1].loadXMLFile(arg)
            #networks[1].topologicalSort()
            #matchBehaviorIds(networks[0],networks[1])
            #print networks[0].toxml()
            #print networks[1].toxml()
            #lcsTree = generalize(networks[0],networks[1])
            #mergedNetwork = merge(lcsTree, networks[0],networks[1])
            #networks[0] = mergedNetwork
            #networks[0].topologicalSort()  #merge already returns a topologically-sorted network
            #networks[1] = BehaviorNetwork()
            
    print mergedNetwork.toxml()
if __name__ == '__main__':
    main(sys.argv[1:])
