#define N_IMPLEMENTS nBoatEngine
//------------------------------------------------------------------------------
//  (C) 2005	Chris Miles
//------------------------------------------------------------------------------
#include "lagoon/nboatengine.h"
#include "enkiduo/colours.h"

nNebulaScriptClass(nBoatEngine, "nphysicalaspect");

//------------------------------------------------------------------------------
nBoatEngine::nBoatEngine() {
}
//------------------------------------------------------------------------------
void nBoatEngine::Initialize(){
    nPhysicalAspect::Initialize();
    //this->AddToDisplayList();
}
//------------------------------------------------------------------------------
void nBoatEngine::InitVars(){
    nPhysicalAspect::InitVars();
    rot.Connect                         ( ent, "rot" );
    vel.Connect                         ( ent, "vel" );
    throttle.Connect                    ( ent, "throttle" );
    rudder.Connect                      ( ent, "rudder" );
    enginePower.Connect                 ( ent, "enginePower" );
    rudderTorque.Connect                ( ent, "rudderTorque" );
    engineRudderDissipation.Connect     ( ent, "engineRudderDissipation" );
    keelForce.Connect                   ( ent, "keelForce" );

    rot.DefaultQ                        ( 0,0,0,1 );
    vel.DefaultV                        ( 0,0,0 );
    throttle.DefaultF                   ( 0.0f );
    rudder.DefaultF                     ( 0.0f );
    enginePower.DefaultF                ( 1.0f );
    rudderTorque.DefaultF               ( 1.0f );
    engineRudderDissipation.DefaultF    ( 0.0f );
    keelForce.DefaultF                  ( 10.0f );

    mass.Connect                        ( ent, "mass" );
    friction.Connect                    ( ent, "friction" );
    //    friction.SetF                       ( 0.0f );
    maxSpeed.Connect                    ( ent, "maxSpeed" );
    maxSpeed.DefaultF                   ( 1.0f );
    accelSpeed.Connect                  ( ent, "accelSpeed" );
    accelSpeed.DefaultF                 ( 1.0f );

    //degradeToCovePhysics.Connect        ( ent, "degradeToCovePhysics" );
    //degradeToCovePhysics.DefaultI       ( 0 );
    //degradedToCove                      = false;

    nEnt* options                       = LOOKUP(nEnt*, "/options" );
    n_assert                            ( options && "failed to find /options || its not an ent" );
    degradeToCovePhysics.Connect              ( options, "cove" );
}
//------------------------------------------------------------------------------
nBoatEngine::~nBoatEngine(){
}
/**-----------------------------------------------------------------------------//
 * boat engine tick - convert throttle / rudder into forces / torques
 * also convert to cove style physics if requested
 * this means i want no friction, but a cap on the maximum speed a boat can reach
 * okay to do this - calculate maximum speed with friction
 * then determine what engine force is required to reach a comparable speed with the other technique
 * then adjust the two parameters accordingly
 * this is a nice tradeoff because it allow us to keep the physics in a good shape, but still behave like cove
 * an alternative technique would be to actually manipulate the force / torque settings we have here.
 * OR we could just directly change velocity
 * well okay for directly changing velocity, we abstract the parameters to a maxSpeed, accelSpeed and turn speed, adn then directly manipulate those
 * its easy to calculate their corresponding values and easy to apply them, thats nice because then the physics is identical
 * should we directly change velocity - like cove would do, or just apply enough force to overcome the friction difference???
 * ets directly change velocity to be most authentic
 * well i think that will fuck things up a lot more, we are better off changing the forces we apply to compensate
--------------------------------------------------------------------------------*/
void nBoatEngine::Tick( float time, float dtime ){
    nPhysicalAspect::Tick( time, dtime );
    n_assert( physics.isvalid() );
    if( mass.isdirty() || enginePower.isdirty() || friction.isdirty() ) {
        //printf("%s calculating maxSpeed\n", str(this));
        maxSpeed.SetF( enginePower.GetF() / ( mass.GetF() * friction.GetF() ) );
        accelSpeed.SetF( maxSpeed.GetF() * (2.0f / friction.GetF()) );
    }

    vector3 v                           = vel.GetV();
    quaternion q                        = rot.GetQ();
    vector3 forward                     = q.rotate(vector3(0, 0,-1));
    vector3 right                       = q.rotate(vector3(1, 0, 0));

    //how much force we are getting from the engine
    float force                         = n_clamp(throttle.GetF(), -1.0f, 1.0f) * enginePower.GetF() * dtime;

    //Use crappy cove style physics - no friction / inertia -----------------------//
    //to do this we scale the force to match their physics model -> since this is the least intrusive way into our system
    //we should divide the force in half, to keep the maxspeeds the same, and then add in enough force to cancel out the friction on the boat
    if ( degradeToCovePhysics.GetI() > 0 && force > 0.01 ){
        vector3 frictionForce           = (-v * friction.GetF() * dtime) * mass.GetF();
        float newForce;
        if( v.len() < throttle.GetF() * maxSpeed.GetF() )
            newForce                    = (force / 2.0f) + frictionForce.len(); //our force is essentially double theirs but it has to contend with friction, so divide in half and subtract friction
        else                            //clamp our maxspeed
            newForce                    = 0;
        force = newForce;
    }

    //calculate keel forces
    float velIntoKeel                   = right % v;
    //float keelF                         = -1.0f * velIntoKeel * keelForce.GetF();
    float keelF                         = -1.0f * velIntoKeel * keelForce.GetF() * dtime;
    //rudder effectiveness should be a summation of engine force and actual velocity
    float inlineVel                     = forward % v;
    float velOverRudder                 = force * (1.0f - engineRudderDissipation.GetF()) + inlineVel;
    //float torque                        = velOverRudder * rudder.GetF() * rudderTorque.GetF();
    float rudderDirection               = n_clamp(rudder.GetF(), -1.0f, +1.0f);
    float torque                        = velOverRudder * rudderDirection * rudderTorque.GetF() * dtime;

    if( this->displayed ) {
        vector3 p = ent->GetVarV("pos");
        this->DrawLine(p, p + 100.0f * forward, white, -1);
        this->DrawLine(p, p + 100.0f * forward * inlineVel, blue);
        this->DrawLine(p, p + 100.0f * right * torque, red);
        this->DrawLine(p, p + 100.0f * right * keelF, green, -30);
        this->DrawLine(p, p + 100.0f * forward * (-1 * velOverRudder), yellow);
    }

    physics->ApplyAlignedForce          (vector3(keelF, 0, -force));
    physics->ApplyAlignedTorque         (vector3(0, torque, 0));
}
